/*
 * This is a module which is used for logging packets.
 */

/* (C) 1999-2001 Paul `Rusty' Russell
 * (C) 2002-2004 Netfilter Core Team <coreteam@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
/*
modification history
--------------------
01a,30Jul09,pkg merged utm fix of spr 11729.

*/

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <net/route.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_LOG.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Netfilter Core Team <coreteam@netfilter.org>");
MODULE_DESCRIPTION("iptables syslog logging module");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

/* 
 * Flag for enabling the logging of packet. It can have the following values:
 * 0 - Nothing logged in INPUT and OUTPUT chains
 * 1 - Log only UNICAST Packets in INPUT and OUTPUT chains
 * 2 - Log only BROADCAST and MULTICAST packets in INPUT and OUTPUT chains
 * 3 - Log UNICAST, BROADCAST and MULTICAST packets in INPUT and OUTPUT chains 
 */
int tf1_log_flag = 0;


/* Use lock to serialize, so printks don't overlap */
static DEFINE_SPINLOCK(log_lock);

#define WAN     "WAN"
#define LAN     "LAN"
#define SELF    "SELF"    
#define LAN_IFACE "eth0.2"

int checkBcastAddr(char * addr)
    {
    if (((unsigned char)addr[0] == 0xff) || ((unsigned char)addr[1] == 0xff) ||
        ((unsigned char)addr[2] == 0xff) || ((unsigned char)addr[3] == 0xff))
        return 1;
    return 0;
    }




int printInterface (const struct net_device *iface,char * logBuff)
    {
    int i = -1;
    if (!iface)
        {
        i = snprintf(logBuff,10,"%s ", SELF);
        return i;
        }

    if (strncmp (iface->name, LAN_IFACE, sizeof(iface->name)) == 0)
        i = snprintf(logBuff,10,"%s ", LAN);
    else
        i = snprintf(logBuff,10,"%s ", WAN);

    return i;
    }



/* One level of recursion won't kill us */
static int dump_packet(const struct nf_loginfo *info,
			const struct sk_buff *skb,
			unsigned int iphoff,char * logBuff)
{
	struct iphdr _iph, *ih;
	unsigned int logflags;
	int len = -1;
	int totLen = 0;

	if (info->type == NF_LOG_TYPE_LOG)
		logflags = info->u.log.logflags;
	else
		logflags = NF_LOG_MASK;

	ih = skb_header_pointer(skb, iphoff, sizeof(_iph), &_iph);
	if (ih == NULL) {
		len = snprintf (logBuff,10,"TRUNCATED");
		totLen += len;
		return totLen;
	}

	/* Important fields:
	 * TOS, len, DF/MF, fragment offset, TTL, src, dst, options. */
	/* Max length: 40 "SRC=255.255.255.255 DST=255.255.255.255 " */
	len = snprintf (logBuff,60,"SRC=%u.%u.%u.%u DST=%u.%u.%u.%u ",
			NIPQUAD(ih->saddr), NIPQUAD(ih->daddr));
	logBuff = logBuff + len;
	totLen += len;

#ifdef FIMPLE
	/* Max length: 46 "LEN=65535 TOS=0xFF PREC=0xFF TTL=255 ID=65535 " */
	printk("LEN=%u TOS=0x%02X PREC=0x%02X TTL=%u ID=%u ",
	       ntohs(ih->tot_len), ih->tos & IPTOS_TOS_MASK,
	       ih->tos & IPTOS_PREC_MASK, ih->ttl, ntohs(ih->id));

	/* Max length: 6 "CE DF MF " */
	if (ntohs(ih->frag_off) & IP_CE)
		printk("CE ");
	if (ntohs(ih->frag_off) & IP_DF)
		printk("DF ");
	if (ntohs(ih->frag_off) & IP_MF)
		printk("MF ");

	/* Max length: 11 "FRAG:65535 " */
	if (ntohs(ih->frag_off) & IP_OFFSET)
		printk("FRAG:%u ", ntohs(ih->frag_off) & IP_OFFSET);

	if ((logflags & IPT_LOG_IPOPT)
	    && ih->ihl * 4 > sizeof(struct iphdr)) {
		unsigned char _opt[4 * 15 - sizeof(struct iphdr)], *op;
		unsigned int i, optsize;

		optsize = ih->ihl * 4 - sizeof(struct iphdr);
		op = skb_header_pointer(skb, iphoff+sizeof(_iph),
					optsize, _opt);
		if (op == NULL) {
			printk("TRUNCATED");
			return;
		}

		/* Max length: 127 "OPT (" 15*4*2chars ") " */
		printk("OPT (");
		for (i = 0; i < optsize; i++)
			printk("%02X", op[i]);
		printk(") ");
	}
#endif /* FIMPLE */    

	switch (ih->protocol) {
	case IPPROTO_TCP: {
		struct tcphdr _tcph, *th;

		/* Max length: 10 "PROTO=TCP " */
		len = snprintf (logBuff,15,"PROTO=TCP ");
		logBuff = logBuff + len;
		totLen += len;

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		th = skb_header_pointer(skb, iphoff + ih->ihl * 4,
					sizeof(_tcph), &_tcph);
		if (th == NULL) {
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
					skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
			break;
		}

		/* Max length: 20 "SPT=65535 DPT=65535 " */
		len = snprintf (logBuff,50,"SPT=%u DPT=%u ",
				ntohs(th->source), ntohs(th->dest));
		logBuff = logBuff + len;
		totLen += len;
#ifdef FIMPLE
		/* Max length: 30 "SEQ=4294967295 ACK=4294967295 " */
		if (logflags & IPT_LOG_TCPSEQ)
			printk("SEQ=%u ACK=%u ",
			       ntohl(th->seq), ntohl(th->ack_seq));
		/* Max length: 13 "WINDOW=65535 " */
		printk("WINDOW=%u ", ntohs(th->window));
		/* Max length: 9 "RES=0x3F " */
		printk("RES=0x%02x ", (u8)(ntohl(tcp_flag_word(th) & TCP_RESERVED_BITS) >> 22));
		/* Max length: 32 "CWR ECE URG ACK PSH RST SYN FIN " */
		if (th->cwr)
			printk("CWR ");
		if (th->ece)
			printk("ECE ");
		if (th->urg)
			printk("URG ");
		if (th->ack)
			printk("ACK ");
		if (th->psh)
			printk("PSH ");
		if (th->rst)
			printk("RST ");
		if (th->syn)
			printk("SYN ");
		if (th->fin)
			printk("FIN ");
		/* Max length: 11 "URGP=65535 " */
		printk("URGP=%u ", ntohs(th->urg_ptr));

		if ((logflags & IPT_LOG_TCPOPT)
		    && th->doff * 4 > sizeof(struct tcphdr)) {
			unsigned char _opt[4 * 15 - sizeof(struct tcphdr)];
			unsigned char *op;
			unsigned int i, optsize;

			optsize = th->doff * 4 - sizeof(struct tcphdr);
			op = skb_header_pointer(skb,
						iphoff+ih->ihl*4+sizeof(_tcph),
						optsize, _opt);
			if (op == NULL) {
				printk("TRUNCATED");
				return;
			}

			/* Max length: 127 "OPT (" 15*4*2chars ") " */
			printk("OPT (");
			for (i = 0; i < optsize; i++)
				printk("%02X", op[i]);
			printk(") ");
		}
#endif /* FIMPLE */
		break;
	}
	case IPPROTO_UDP: {
		struct udphdr _udph, *uh;

		/* Max length: 10 "PROTO=UDP " */
		len = snprintf (logBuff,15,"PROTO=UDP ");
		logBuff = logBuff + len;
		totLen += len;

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		uh = skb_header_pointer(skb, iphoff+ih->ihl*4,
					sizeof(_udph), &_udph);
		if (uh == NULL) {
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
					skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
			break;
		}
#ifdef FIMPLE	
		/* Max length: 20 "SPT=65535 DPT=65535 " */
		printk("SPT=%u DPT=%u LEN=%u ",
		       ntohs(uh->source), ntohs(uh->dest),
		       ntohs(uh->len));
#else

		/* Max length: 20 "SPT=65535 DPT=65535 " */
		len = snprintf (logBuff,50,"SPT=%u DPT=%u",
				ntohs(uh->source), ntohs(uh->dest));
		logBuff = logBuff + len;
		totLen += len;
#endif /* FIMPLE */
		break;
	}
	case IPPROTO_ICMP: {
		struct icmphdr _icmph, *ich;
#ifdef FIMPLE		
		static const size_t required_len[NR_ICMP_TYPES+1]
			= { [ICMP_ECHOREPLY] = 4,
			    [ICMP_DEST_UNREACH]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_SOURCE_QUENCH]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_REDIRECT]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_ECHO] = 4,
			    [ICMP_TIME_EXCEEDED]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_PARAMETERPROB]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_TIMESTAMP] = 20,
			    [ICMP_TIMESTAMPREPLY] = 20,
			    [ICMP_ADDRESS] = 12,
			    [ICMP_ADDRESSREPLY] = 12 };
#endif /* FIMPLE */
		/* Max length: 11 "PROTO=ICMP " */
		len = snprintf (logBuff,15,"PROTO=ICMP ");
		logBuff = logBuff + len;
		totLen += len;

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		ich = skb_header_pointer(skb, iphoff + ih->ihl * 4,
					 sizeof(_icmph), &_icmph);
		if (ich == NULL) {
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
					skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
			break;
		}

		/* Max length: 18 "TYPE=255 CODE=255 " */
		len = snprintf (logBuff,60,"TYPE=%u CODE=%u ",
				ich->type, ich->code);
		logBuff = logBuff + len;
		totLen += len;

#ifdef FIMPLE		
		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		if (ich->type <= NR_ICMP_TYPES
		    && required_len[ich->type]
		    && skb->len-iphoff-ih->ihl*4 < required_len[ich->type]) {
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			break;
		}

		switch (ich->type) {
		case ICMP_ECHOREPLY:
		case ICMP_ECHO:
			/* Max length: 19 "ID=65535 SEQ=65535 " */
			printk("ID=%u SEQ=%u ",
			       ntohs(ich->un.echo.id),
			       ntohs(ich->un.echo.sequence));
			break;

		case ICMP_PARAMETERPROB:
			/* Max length: 14 "PARAMETER=255 " */
			printk("PARAMETER=%u ",
			       ntohl(ich->un.gateway) >> 24);
			break;
		case ICMP_REDIRECT:
			/* Max length: 24 "GATEWAY=255.255.255.255 " */
			printk("GATEWAY=%u.%u.%u.%u ",
			       NIPQUAD(ich->un.gateway));
			/* Fall through */
		case ICMP_DEST_UNREACH:
		case ICMP_SOURCE_QUENCH:
		case ICMP_TIME_EXCEEDED:
			/* Max length: 3+maxlen */
			if (!iphoff) { /* Only recurse once. */
				printk("[");
				dump_packet(info, skb,
					    iphoff + ih->ihl*4+sizeof(_icmph));
				printk("] ");
			}

			/* Max length: 10 "MTU=65535 " */
			if (ich->type == ICMP_DEST_UNREACH
			    && ich->code == ICMP_FRAG_NEEDED)
				printk("MTU=%u ", ntohs(ich->un.frag.mtu));
		}
#endif /* FIMPLE */
		break;
	}
	/* Max Length */
	case IPPROTO_AH: {
		struct ip_auth_hdr _ahdr, *ah;

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;
		
		/* Max length: 9 "PROTO=AH " */
		len = snprintf (logBuff,10,"PROTO=AH ");
		logBuff = logBuff + len;
		totLen += len; 

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		ah = skb_header_pointer(skb, iphoff+ih->ihl*4,
					sizeof(_ahdr), &_ahdr);
		if (ah == NULL) {
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
					skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
			break;
		}

		/* Length: 15 "SPI=0xF1234567 " */
		len = snprintf (logBuff,30,"SPI=0x%x ", ntohl(ah->spi));
		logBuff = logBuff + len;
		totLen += len;
		break;
	}
	case IPPROTO_ESP: {
		struct ip_esp_hdr _esph, *eh;

		/* Max length: 10 "PROTO=ESP " */
		len = snprintf (logBuff,15,"PROTO=ESP ");
		logBuff = logBuff + len;
		totLen += len; 

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		eh = skb_header_pointer(skb, iphoff+ih->ihl*4,
					sizeof(_esph), &_esph);
		if (eh == NULL) {
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
					skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
			break;
		}

		/* Length: 15 "SPI=0xF1234567 " */
		len = snprintf (logBuff,30,"SPI=0x%x ", ntohl(eh->spi));
		logBuff = logBuff + len;
		totLen += len;
		break;
	}
	/* Max length: 10 "PROTO 255 " */
	default:
		len = snprintf (logBuff,15,"PROTO=%u ", ih->protocol);
		logBuff = logBuff + len;
		totLen += len; 
	}

	/* Max length: 15 "UID=4294967295 " */
 	if ((logflags & IPT_LOG_UID) && !iphoff && skb->sk) {
		read_lock_bh(&skb->sk->sk_callback_lock);
		if (skb->sk->sk_socket && skb->sk->sk_socket->file)
 			printk("UID=%u ", skb->sk->sk_socket->file->f_uid);
		read_unlock_bh(&skb->sk->sk_callback_lock);
	}

	return totLen;
	/* Proto    Max log string length */
	/* IP:      40+46+6+11+127 = 230 */
	/* TCP:     10+max(25,20+30+13+9+32+11+127) = 252 */
	/* UDP:     10+max(25,20) = 35 */
	/* ICMP:    11+max(25, 18+25+max(19,14,24+3+n+10,3+n+10)) = 91+n */
	/* ESP:     10+max(25)+15 = 50 */
	/* AH:      9+max(25)+15 = 49 */
	/* unknown: 10 */

	/* (ICMP allows recursion one level deep) */
	/* maxlen =  IP + ICMP +  IP + max(TCP,UDP,ICMP,unknown) */
	/* maxlen = 230+   91  + 230 + 252 = 803 */
}

static struct nf_loginfo default_loginfo = {
	.type	= NF_LOG_TYPE_LOG,
	.u = {
		.log = {
			.level    = 0,
			.logflags = NF_LOG_MASK,
		},
	},
};


#define IS_PKT_UNICAST(x,y)                         \
    ((((inet_addr_type(x) == RTN_UNICAST) ||        \
       (inet_addr_type(x) == RTN_LOCAL))) &&        \
     (((inet_addr_type(y) == RTN_UNICAST)) ||       \
      ((inet_addr_type(y) == RTN_LOCAL))))   


void dump_packet_wrapper(const struct ipt_log_info *info,
			const struct sk_buff *skb,
			unsigned int iphoff)
    {
    char * pTmp = NULL;    
    char logBuff[1024];
    int len = 0;
    struct nf_loginfo nfloginfo;

    nfloginfo.u.log.level = info->level;
    nfloginfo.u.log.logflags = info->logflags;
    pTmp = logBuff;
    len = snprintf (pTmp,40,"%s ",info->prefix);
    pTmp = pTmp + len;    
    dump_packet(&nfloginfo, skb, iphoff,pTmp);
    printk("%s\n",logBuff);
    return;
    }

static int
ipt_log_packet(unsigned int pf,
	       unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const struct ipt_log_info *loginfo,
	       const char *prefix,
	       char * logBuff)
{
	
	/*if (!loginfo)
		loginfo = &default_loginfo;

	spin_lock_bh(&log_lock);
	printk("<%d>%sIN=%s OUT=%s ", loginfo->u.log.level,
	       prefix,
	       in ? in->name : "",
	       out ? out->name : "");*/

	struct iphdr _iph,*ih;
	int len = -1;
	int totLen = 0;
	struct nf_loginfo nfloginfo;

    
	spin_lock_bh(&log_lock);
    
	ih = skb_header_pointer(skb, 0, sizeof(_iph), &_iph);
	if (ih == NULL) 
        {
        spin_unlock_bh(&log_lock);
        return 1;
        }

#ifdef FIMPLE
	if (checkBcastAddr((char *)&ih->saddr) ||
        checkBcastAddr((char *)&ih->daddr)) {
        spin_unlock_bh(&log_lock);
        return 1;
	}
#endif /* FIMPLE */
    
    /* Check whether the packet is for INPUT, OUTPUT or FORWARD chain */
    switch (hooknum)
        {
        case NF_IP_LOCAL_IN:
        case NF_IP_LOCAL_OUT:
        {
        /* Check the value of proc entry tf1_log_flag */
        switch ( tf1_log_flag )
            {
            case IPT_LOG_UNICAST:
            {
            /* Log only UNICAST packets in INPUT and OUTPUT chain */
            
            /* 
             * Checking whether the packets source or destination
             * address is UNICAST or LOCAL. Here LOCAL means ip
             * address of the box
             */
            if (!(IS_PKT_UNICAST(ih->saddr,ih->daddr)))
                {
                spin_unlock_bh(&log_lock);
                return 1;
                }
            
            /*
             * Checking local generated packets.
             * Example:127.0.0.1<->127.0.0.1
             */
            if (LOOPBACK(ih->saddr) || LOOPBACK(ih->daddr) )
                {
                spin_unlock_bh(&log_lock);
                return 1;
                }
            break;
            }
            case IPT_LOG_BCAST_MCAST:
            {
            /* Checking whether the packet is UNICAST or LOCAL */
            if (IS_PKT_UNICAST(ih->saddr, ih->daddr))
                {
                spin_unlock_bh(&log_lock);
                return 1;
                }
            break;
            }
            case IPT_LOG_UCAST_BCAST_MCAST:
            {
            /* 
             * Checking local generated packets.
             * Example:127.0.0.1<->127.0.0.1
             */
            if (LOOPBACK(ih->saddr) || LOOPBACK(ih->daddr) )
                {
                spin_unlock_bh(&log_lock);
                return 1;
                }

            /* 
             * Log UNICAST,MULTICAST and BROADCAST packets in INPUT
             * and OUTPUT chains
             */
            break;
            }
            case IPT_LOG_NO_PKT:
            default: 
            {
#ifdef CONFIG_ARCH_NG_FVS336GV2
            if(strcmp(loginfo->prefix,"WAN2LAN[DROP]") == 0 ||
 		strcmp(loginfo->prefix,"WAN2DMZ[DROP]") == 0)
#else /* CONFIG_ARCH_NG_FVS336GV2 */
            if(strcmp(loginfo->prefix,"WAN2LAN[DROP]") == 0)
#endif /* !CONFIG_ARCH_NG_FVS336GV2 */       
                break;

            /* 
             * Nothing to be logged, this is the scenario when
             * tf1_log_flag value is other than 0,1,2,3. 
             */
            
            spin_unlock_bh(&log_lock);
            return 1;
            }
            } /* End of Inner switch*/ 
        break ;
        } /* End of outer case */
        
        default:
        {
        /* 
         * All the packets should be logged except coming from
         * INPUT and OUTPUT chains which have rule configured
         * for them.
         */
        
        /* Checking the broadcast packets */
        if ((inet_addr_type(ih->saddr) == RTN_BROADCAST) ||
            ((inet_addr_type(ih->daddr) == RTN_BROADCAST)))
            {
            spin_unlock_bh(&log_lock);
            return 1;
            }
        }
        
        } /* End of outer switch */
    
    /* print the log values in syslog */ 
   // printk(level_string);
    
#ifdef FIMPLE    
	printk("%sIN=%s OUT=%s ",
	       prefix == NULL ? loginfo->prefix : prefix,
	       in ? in->name : "", out ? out->name : "");
#endif /* FIMPLE */
	len = snprintf (logBuff,5," IN=");
	logBuff = logBuff + len;
	totLen += len;
	len = printInterface(in,logBuff);
	logBuff = logBuff + len;
	totLen += len;
    
	len = snprintf (logBuff,6," OUT=");
	logBuff = logBuff + len;
	totLen += len;
	len = printInterface(out,logBuff);
	logBuff = logBuff + len;
	totLen += len;

#ifdef CONFIG_BRIDGE_NETFILTER
	if (skb->nf_bridge) {
		struct net_device *physindev = skb->nf_bridge->physindev;
		struct net_device *physoutdev = skb->nf_bridge->physoutdev;

		if (physindev && in != physindev)
			printk("PHYSIN=%s ", physindev->name);
		if (physoutdev && out != physoutdev)
			printk("PHYSOUT=%s ", physoutdev->name);
	}
#endif

#ifdef FIMPLE
	if (in && !out) {
		/* MAC logging for input chain only. */
		printk("MAC=");
		if (skb->dev && skb->dev->hard_header_len
		    && skb->mac.raw != (void*)skb->nh.iph) {
			int i;
			unsigned char *p = skb->mac.raw;
			for (i = 0; i < skb->dev->hard_header_len; i++,p++)
				printk("%02x%c", *p,
				       i==skb->dev->hard_header_len - 1
				       ? ' ':':');
		} else
			printk(" ");
	}
#endif /*FIMPLE*/
	
	nfloginfo.u.log.level = loginfo->level;
	nfloginfo.u.log.logflags = loginfo->logflags;

	len = dump_packet(&nfloginfo, skb, 0,logBuff);
	logBuff = logBuff + len;
	totLen += len;

	spin_unlock_bh(&log_lock);
    return 0;
}

static unsigned int
ipt_log_target(struct sk_buff **pskb,
	       const struct net_device *in,
	       const struct net_device *out,
	       unsigned int hooknum,
	       const void *targinfo,
	       void *userinfo)
{
	const struct ipt_log_info *loginfo = targinfo;
	struct nf_loginfo li;
	char logBuff[1024];
	char * pTmp = logBuff;
	char * pFormat = NULL;
	int len = 0;
	int tokenFlag = 0;
	int formatFlag = 0;
	unsigned char *p = (*pskb)->mac.raw;
    int retVal = 0;

    pFormat = (char *) loginfo->format;
//	level_string[1] = '0' + (loginfo->level % 8);

    len = snprintf (pTmp,40,"%s ",loginfo->prefix);
    pTmp = pTmp + len;

    if (userinfo)
        {
        len = snprintf (pTmp, 128 ,"%s ",userinfo);
        pTmp = pTmp + len;
        }

    while((*pFormat) != '\0')
    {
        tokenFlag = 0;
        formatFlag = 0;
        if(*pFormat == '%')
        {   
            switch(*(++pFormat))
            {
                case '.': 
                    formatFlag = 1;     
                    ++pFormat;
                    break;    
            }    

            switch(*pFormat)
            {
                case 's':
                    tokenFlag = 1;   
                    ++pFormat;
                    break;
                case 'd':
                    tokenFlag = 2;
                    ++pFormat;
                    break;    
            }

            switch(*(pFormat))
            {
                case 'i':
                    break;     
                case 'm':
                    if(tokenFlag != 2)
                    {
                        len = snprintf (pTmp,100,"SRC MAC = %02x:%02x:%02x:%02x"
                                ":%02x:%02x ",*(p+6),*(p+7),*(p+8),*(p+9),
                                *(p+10),*(p+11));
                        pTmp = pTmp + len;
                    }
                    else
                    {
                        len = snprintf (pTmp,100,"DST MAC = %02x:%02x:%02x:%02x"
                                ":%02x:%02x ",*p,*(p+1),*(p+2),*(p+3),
                                *(p+4),*(p+5));
                        pTmp = pTmp + len;
                    }
                    break;     
                case 'p':
                    break;     
                case 'u':
                    break;     
                case 'I':
                    break;     
                case 'T':
                    break;     
                case 'U':
                    break;     
                case 'C':
                    break;     
                case 'S':
                    break;     
                case 'L':
                    len = snprintf (pTmp,250,"ETHER[%02x:%02x:%02x:%02x:%02x:"
                            "%02x -> %02x:%02x:%02x:%02x:%02x:%02x ]",*(p+6),
                            *(p+7),*(p+8),*(p+9),*(p+10),*(p+11),*p,
                            *(p+1),*(p+2),*(p+3),*(p+4),*(p+5));
                    pTmp = pTmp + len;
                    break;    
                case 'P':
                    retVal = ipt_log_packet(PF_INET, hooknum, *pskb, in, out, loginfo, 
                                   NULL,pTmp);
                     break;
                default :
                    if(tokenFlag != 0)
                        len = snprintf(pTmp,5,"%%%c%c",*(pFormat-1),*pFormat);
                    else
                        len = snprintf(pTmp,5,"%%%c",*pFormat);
                    pTmp = pTmp + len;
            }    
        }
        else
        {
            len = snprintf (pTmp,5,"%c",*(pFormat));
            pTmp = pTmp + len;
        }
        pFormat++;
    }
	if (strlen(loginfo->format) == 0)
		{
		if (loginfo->logflags & IPT_LOG_NFLOG)
			{
			li.type = NF_LOG_TYPE_LOG;
			li.u.log.level = loginfo->level;
			li.u.log.logflags = loginfo->logflags;
			nf_log_packet(PF_INET, hooknum, *pskb, in, out, &li,
				      loginfo->prefix);
            retVal = 0;
			}
		else
			retVal = ipt_log_packet(PF_INET, hooknum, *pskb, in, out, loginfo,
				       loginfo->prefix, pTmp);
		}
    if(retVal == 0)
        printk("%s\n",logBuff);

	return IPT_CONTINUE;
}

int ipt_log_target_ex(struct sk_buff **pskb,
	       const struct net_device *in,
	       const struct net_device *out,
	       unsigned int hooknum,
	       const void *targinfo,
	       void *userinfo)
    {
    return  ipt_log_target(pskb,in,out,hooknum,targinfo,userinfo);
    }

void
ipt_logfn( unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const char *prefix)

{

	struct ipt_log_info loginfo = { 
		.level = 4, /* WARNING */ 
		.logflags = 0, 
		.prefix = "", 
		.format = "" 
	};

//		ipt_log_packet(PF_INET, hooknum, skb, in, out, loginfo,
//		               loginfo->prefix, logBuff);
    /* 
     * the prefix is sent as userinfo argument, because it includes the prefix
     * and the url information 
     */ 
	ipt_log_target(&skb, in, out, hooknum, &loginfo, prefix);
}		

static void nf_log_packet_wrapper (unsigned int pf,
	       unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const struct nf_loginfo *nfloginfo,
	       const char *prefix)
{
        char logBuff [1024];
	struct ipt_log_info info;


	info.level = nfloginfo->u.log.level;
	info.logflags = nfloginfo->u.log.logflags;
        ipt_log_packet (pf, hooknum, skb, in, out, &info, prefix, logBuff);
	
}
	
static int ipt_log_checkentry(const char *tablename,
			      const void *e,
			      void *targinfo,
			      unsigned int targinfosize,
			      unsigned int hook_mask)
{
	const struct ipt_log_info *loginfo = targinfo;

	if (targinfosize != IPT_ALIGN(sizeof(struct ipt_log_info))) {
		DEBUGP("LOG: targinfosize %u != %u\n",
		       targinfosize, IPT_ALIGN(sizeof(struct ipt_log_info)));
		return 0;
	}

	if (loginfo->level >= 8) {
		DEBUGP("LOG: level %u >= 8\n", loginfo->level);
		return 0;
	}

	if (loginfo->prefix[sizeof(loginfo->prefix)-1] != '\0') {
		DEBUGP("LOG: prefix term %i\n",
		       loginfo->prefix[sizeof(loginfo->prefix)-1]);
		return 0;
	}

	if (loginfo->format[sizeof(loginfo->format)-1] != '\0') {
		DEBUGP("LOG: format term %i\n",
		       loginfo->format[sizeof(loginfo->format)-1]);
		return 0;
	}

	return 1;
}

static struct ipt_target ipt_log_reg = {
	.name		= "LOG",
	.target		= ipt_log_target,
	.checkentry	= ipt_log_checkentry,
	.me		= THIS_MODULE,
};

static struct nf_logger ipt_log_logger ={
	.name		= "ipt_LOG",
	.logfn		= &nf_log_packet_wrapper,
	.me		= THIS_MODULE,
};

static int __init init(void)
{
	if (ipt_register_target(&ipt_log_reg))
		return -EINVAL;
	if (nf_log_register(PF_INET, &ipt_log_logger) < 0) {
		printk(KERN_WARNING "ipt_LOG: not logging via system console "
		       "since somebody else already registered for PF_INET\n");
		/* we cannot make module load fail here, since otherwise
		 * iptables userspace would abort */
	}
	
	return 0;
}

static void __exit fini(void)
{
	nf_log_unregister_logger(&ipt_log_logger);
	ipt_unregister_target(&ipt_log_reg);
}

module_init(init);
module_exit(fini);
